# This was hacked from the uparrow package written by Tom Holroyd
# Right now its a mess. :)

namespace eval term {
    namespace export save_pos num_rows

   # line is a list of the
    # characters on the current line.  point is where the cursor is,
    # and event is the current history event.

    array set term {
        line    {}
        point   0
        event   1
	clearonhit 0
    }

    # Variables used by some of the functions.

    # These are the default functions to call when keys are pressed.
    # Esc has subcommands.

    array set bindings {
        "\x04"      exit
	"\x01"      control-a
        "\x05"      control-e
        "\x15"      control-u
	"\x1B"      esc
        "\x1B\[A"   up
        "\x1B\[B"   down
        "\x1B\[C"   right
        "\x1B\[D"   left
        "\x1BOA"    up
        "\x1BOB"    down
        "\x1BOC"    right
        "\x1BOD"    left
    }

    if {$env(TERM) == "xterm"} {
	array set bindings {
	    "\x7F"      delch
	    "\x08"      backspace
	}
    } elseif {$env(TERM) == "linux"} {
	array set bindings {
	    "\x7F"      backspace
	    "~"         delch
	}
    }

    # The linux delete key isn't right
}


proc term::save_pos {} {
    puts -nonewline \0337
}

proc term::restore_pos {} {
    puts -nonewline \0338
}

proc term::goto_rowcol {row col} {
    puts -nonewline \033\[${row}\;${col}f
}

proc term::erase_screen {} {
    puts -nonewline \033\[2J
}

proc term::erase_toeol {} {
    puts -nonewline \033\[K
}

proc term::reset {} {
    puts -nonewline \033c
}

proc term::scroll_region {top bottom} {
    puts -nonewline \033\[${top}\;${bottom}r
}

proc term::move_right {{n 1}} {

}

proc term::move_left {{n 1}} {

}

proc term::num_rows {} {
    global env

    set rows 24
    if {[info exists env(LINES)]} {
	set rows $env(LINES)
    } else {
	catch {set rows [lindex [exec stty size] 0]}
    }
    
    return $rows
}

# cbreak on|off --
# Turn one-character-at-a-time mode on or off.  Unix only.

proc term::cbreak {on} {
    if {$on == "on"} {
	exec stty -icanon -echo min 1 <@ stdin
    } else {
	exec stty icanon echo min 4 <@ stdin
    }
}

########################################################################
# These are the functions that are bound to the various key sequences.

# The default action in case there is no binding is to insert the character.

proc term::insert {c} {
    variable term

    puts -nonewline "\x1B\[@$c"
    set term(line) [linsert $term(line) $term(point) $c]
    incr term(point)

    return
}

# Control-a moves to the beginning of the line.

proc term::control-a {} {
    variable term

    while {$term(point) > 0} {
        puts -nonewline "\x08"
        incr term(point) -1
    }

    return
}

# Control-e moves to the end of the line.

proc term::control-e {} {
    variable term

    set l [llength $term(line)]
    while {$term(point) < $l} {
        puts -nonewline "\x1B\[C"
        incr term(point)
    }

    return
}

# Backspace erases the character to the left of the point and moves left.

proc term::backspace {} {
    variable term

    if {0 < $term(point)} {
        puts -nonewline "\x08\x1B\[P"
        incr term(point) -1
        set term(line) [lreplace $term(line) $term(point) $term(point)]
    }

    return
}

# Delch deletes the character to the right of the point.

proc term::delch {} {
    variable term

    if {$term(point) < [llength $term(line)]} {
        puts -nonewline "\x1B\[P"
        set term(line) [lreplace $term(line) $term(point) $term(point)]
    }

    return
}

# Control-u erases the entire command line.

proc term::control-u {} {
    variable term

    while {$term(point) > 0} {
        puts -nonewline "\x08"
        incr term(point) -1
    }
    erase_toeol

    set term(line) {}

    return
}

# Esc begins a two-character subcommand.

proc term::esc {} {
    variable bindings

    set code "\x1B[read stdin 2]"
    if {[info exists bindings($code)]} {
        $bindings($code)
    }

    return
}

# Right and left just move the insertion point around.

proc term::right {} {
    variable term

    if {$term(point) < [llength $term(line)]} {
        puts -nonewline "\x1B\[C"
        incr term(point) 1
    }

    return
}

proc term::left {} {
    variable term

    if {0 < $term(point)} {
        puts -nonewline "\x1B\[D"
        incr term(point) -1
    }

    return
}

# Up and down move through the history list.

proc term::up {} {
    variable term
    variable saveCmd

    # If we're leaving event 1 (the new command) save it.

    if {$term(event) == 1} {
        set saveCmd [join $term(line) ""]
    }

    # Erase this command and get the previous event.

    control-u
    if {[expr {1 - [history keep]}] < $term(event)} {
        incr term(event) -1
    }
    set line [history event $term(event)]
    puts -nonewline $line
    set term(line) [split $line {}]
    set term(point) [llength $term(line)]

    return
}

proc term::down {} {
    variable term
    variable saveCmd

    # Erase this command and get the next event.

    if {$term(event) == 1} {
        return
    }

    control-u
    incr term(event)
    if {$term(event) == 1} {
        set line $saveCmd
    } else {
        set line [history event $term(event)]
    }
    puts -nonewline $line
    set term(line) [split $line {}]
    set term(point) [llength $term(line)]

    return
}

proc term::endline {} {
    variable term

    puts -nonewline \n
    set term(line) {}
    set term(point) 0
    term::erase_toeol
}

proc term::handle_input {proc} {
    variable term
    variable bindings

    set c [read stdin 1]   

    if {$c == "\x0A"} {
	$proc [join $term(line) ""]
	return
    }
    
    if {[info exists bindings($c)]} {
	$bindings($c)
    } else {
	if {$term(clearonhit)} {
	    endline
	}
	insert $c
    }

    set term(clearonhit) 0
}
